const fs = require('fs');
var ONE_MINUTE = 60 * 1000;

var bytesRead;
var buffer4 = new Buffer(4);

const folderName = '/HCBv2/data/hcb_rrd';

var datFiles = new Array();
var datFileProperties = new Array();
var minuteArray = new Array();
var allArray = new Array();

// Get all the filenames in the database folder
console.log("Getting list of data files");
var fileNames = fs.readdirSync(folderName);

console.log("Creating data file array");
// Copy the .dat files to a seperate array
for (var i = 0, len = fileNames.length; i < len; i++) {
  if (fileNames[i].substr(-4) == ".dat") {
    datFiles.push(folderName + "/" + fileNames[i]);
  }
}

// fileNames array is not used after this, see if this will free up memory
fileNames = null;

console.log("Creating minute array");
// Read the .dat and set function timeout to read the data from it
for (var i = 0, len = datFiles.length; i < len; i++) {
  var properties = getConfig(datFiles[i]);
  datFileProperties[i] = properties;
  if (properties[2] == 60) {
    // Updated every minute
    minuteArray.push(i);
  }
  allArray.push(i);
}

function repeatEvery(func, interval, whichArray) {
    // Check current time and calculate the delay until next interval
    var now = new Date(),
      delay = interval - now % interval;

    delay = delay;
    function start() {
        // Execute function now...
        func(whichArray);
        // ... and every interval
        setInterval(func, interval, whichArray);
    }

    // Delay execution until it's an even interval
    setTimeout(start, delay);
}

console.log("Connecting to mqtt");
var mqtt    = require('mqtt');
var client  = mqtt.connect('mqtt://192.168.176.190', {port: 9032});

client.on('connect', function () {
  console.log("MQTT connected");
  client.publish('presence', 'Greetings from Toon');
  console.log("Invoke repeat function");
  repeatEvery(getMinuteData, ONE_MINUTE, minuteArray);
  // TEST: Get ALL entities, not just the minute ones
  // repeatEvery(getMinuteData, ONE_MINUTE, allArray);
});


// Not in use atm
client.on('message', function (topic, message) {
  // message is Buffer
  console.log(message.toString());
  client.end();
});

// getMinuteData(minuteArray);

function getConfig(fileName) {
  // We're going to return an array with properties
  var properties = new Array();

  var contents = fs.readFileSync(fileName);
  // Skip magic
  pointer = 17;

  // skip deviceUuid
  var decimal = contents.readInt32LE(pointer);
  pointer = pointer + 4 + decimal;

  decimal = contents.readInt32LE(pointer);
  pointer = pointer + 4;

  properties.push(contents.slice(pointer, pointer + decimal - 1)); // deviceVar
  pointer = pointer + decimal;

  decimal = contents.readInt32LE(pointer);
  pointer = pointer + 4;

  //var deviceSvc = contents.slice(pointer, pointer + decimal);
  pointer = pointer + decimal;

  decimal = contents.readInt32LE(pointer);
  pointer = pointer + 4;

  var sampleType = contents.slice(pointer, pointer + decimal - 1);
  properties.push(sampleType);
  pointer = pointer + decimal;

  var isDouble = false;
  var isInt = false;

  if (sampleType.toString() == "double") {
    isDouble = true;
  }
  if (sampleType.toString() == "integer") {
    isInt = true;
  }

  while (pointer < contents.length) {
    if (isDouble) {
      // var doubleValue = contents.readDoubleLE(pointer);
      // console.log("Value: " + doubleValue);
      pointer = pointer + 8;

      // var doubleDivider = contents.readDoubleLE(pointer);
      // console.log("Divider: " + doubleDivider);
      pointer = pointer + 20; // Skip the timestamps and minSamplesPerBin
    }
    if (isInt) {
      // Skip three 4 byte ints that we don't know the function of
      // Then skip timestamps and minSamplesPerBin
      pointer = pointer + 24;
      console.log("Integer");
    }

    decimal = contents.readInt32LE(pointer);
    pointer = pointer + 4;

    properties.push(contents.slice(pointer, pointer + decimal -1)); // binLength, Do not read closing 0
    console.log("binLength: " + contents.slice(pointer, pointer + decimal -1));
    pointer = pointer + decimal + 4; // skip fileOffset

    properties.push(contents.readInt32LE(pointer)); // numSamples

    pointer = pointer + 8; // Skip unkown value

    decimal = contents.readInt32LE(pointer);
    pointer = pointer + 4;

    properties.push(contents.slice(pointer, pointer + decimal - 1)); // Name, leave out the closing zero
    if (isInt) {
      console.log("Name: " + contents.slice(pointer, pointer + decimal -1))
    }
    pointerpointer = pointer + decimal;

    decimal = contents.readInt32LE(pointer);
    pointer = pointer + 4;

    properties.push(contents.slice(pointer, pointer + decimal - 1)); // Consolidator, no closing 0
    pointer = pointer + decimal;
  }

  return properties;
}

function getMinuteData(elements) {
  // Todo: We can get certain variables from existing arrays instead of
  //       reading them from file
  for (var currentElement = 0, len = elements.length; currentElement < len; currentElement++) {
    // console.log("Starting element " + elements[currentElement] + ", file: " + datFiles[elements[currentElement]]);

    var contents = fs.readFileSync(datFiles[elements[currentElement]]);
    // Skip magic
    pointer = 17;

    // skip deviceUuid
    var decimal = contents.readInt32LE(pointer);
    pointer = pointer + 4 + decimal;

    decimal = contents.readInt32LE(pointer);
    pointer = pointer + 4;

    var deviceVar = contents.slice(pointer, pointer + decimal - 1); // deviceVar
    pointer = pointer + decimal;

    decimal = contents.readInt32LE(pointer);
    pointer = pointer + 4;

    //var deviceSvc = contents.slice(pointer, pointer + decimal);
    pointer = pointer + decimal;

    decimal = contents.readInt32LE(pointer);
    pointer = pointer + 4;

    var sampleType = contents.slice(pointer, pointer + decimal - 1);
    // properties.push(sampleType);
    pointer = pointer + decimal;

    var isDouble = false;
    var isInt = false;

    if (sampleType.toString() == "double") {
      isDouble = true;
    }
    if (sampleType.toString() == "integer") {
      isInt = true;
    }

    while (pointer < contents.length) {

      if (isDouble) {
        // var doubleValue = contents.readDoubleLE(pointer);
        // console.log("Value: " + doubleValue);
        pointer = pointer + 8;

        // var doubleDivider = contents.readDoubleLE(pointer);
        // console.log("Divider: " + doubleDivider);
        pointer = pointer + 8;
      }
      if (isInt) {
        // Skip three 4 byte ints that we don't know the function of
        pointer = pointer + 12;
      }

      var timeStamp = contents.readInt32LE(pointer) + 3600;
      var date = new Date(timeStamp*1000);
      var hours = date.getHours();
      var minutes = "0" + date.getMinutes();
      var seconds = "0" + date.getSeconds();
      var formattedTime = hours + ':' + minutes.substr(-2) + ':' + seconds.substr(-2);
      pointer = pointer + 12; // Skip the next timeStamp and minSamplesPerBin

      decimal = contents.readInt32LE(pointer);
      pointer = pointer + 4;

      var binLength = contents.slice(pointer, pointer + decimal -1);
      pointer = pointer + decimal;

      var fileOffset = contents.readInt32LE(pointer);
      pointer = pointer + 4;

      var numSamples = contents.readInt32LE(pointer);
      pointer = pointer + 8; // Skip unkown value

      decimal = contents.readInt32LE(pointer);
      pointer = pointer + 4;

      var name = contents.slice(pointer, pointer + decimal - 1);
      pointer = pointer + decimal;

      decimal = contents.readInt32LE(pointer);
      pointer = pointer + 4;

      var consolidator = contents.slice(pointer, pointer + decimal - 1);
      pointer = pointer + decimal;

      var rraFileName = datFiles[elements[currentElement]].slice(0, -4) + "-" + name + ".rra";
      // console.log("Filename: " + fileName + ".dat");
      // console.log("RRa     : " + rraFileName);

      var rraFile = fs.openSync(rraFileName, 'r');

      if (isDouble) {
        var buffer8 = new Buffer(8);
        // The last value is not at fileoffset * 8, but 8 bytes before that
        bytesRead = fs.readSync(rraFile, buffer8, 0, 8, (fileOffset * 8) - 8);
        console.log("Time: " + formattedTime + ", deviceVar: " + deviceVar + ", value: " + buffer8.readDoubleLE(0) + ", name: " + name);
        if (name == "30days") {
          client.publish("toon", "{\"" + deviceVar + "_" + name + "\": " + buffer8.readDoubleLE(0) + "}");
        } else {
          client.publish("toon", "{\"" + deviceVar + "_" + name + "\": " + buffer8.readDoubleLE(0) + "}");
        }
        // Read from five entries before last
        var tmpOffset = (fileOffset * 8) - (5 * 8);
        for (var i = 0; i < 5; i++) {
          bytesRead = fs.readSync(rraFile, buffer8, 0, 8, tmpOffset + (i * 8));
          // console.log("Read values: " + buffer8.readDoubleLE(0));
        }
      } else if (isInt) {
        var buffer4 = new Buffer(4);
        // The last value is not at fileoffset * 8, but 8 bytes before that
        var bytesRead = fs.readSync(rraFile, buffer4, 0, 4, (fileOffset * 4) - 4);
        var intValue = buffer4.readInt32LE(0);
        if (intValue == 0x7fffffff) {
          console.log ("Infinite value detected");
        }
        intValue = intValue / 1000;
        console.log("Time: " + formattedTime + ", deviceVar: " + deviceVar + ", value: " + intValue + ", name: " + name);
        if (name == "30days") {
          client.publish("toon", "{\"" + deviceVar + "_" + name + "\": " + buffer8.readDoubleLE(0) + "}");
        } else {
          client.publish("toon", "{\"" + deviceVar + "_" + name + "\": " + buffer8.readDoubleLE(0) + "}");
        }
        // Read from five entries before last
        var tmpOffset = (fileOffset * 8) - (5 * 8);
        for (var i = 0; i < 5; i++) {
          bytesRead = fs.readSync(rraFile, buffer8, 0, 8, tmpOffset + (i * 8));
          // console.log("Read values: " + buffer8.readDoubleLE(0));
        }
      } else {
        console.log("Unkown sampleType: " + sampleType);
      }
      fs.closeSync(rraFile);
    }
  }
}
